import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.macro.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import ij.plugin.*;
import ij.plugin.frame.*;
import ij.measure.*;
import ij.util.Tools;
import java.awt.datatransfer.*;

public class Interactive_Fitting implements PlugIn, ImageListener {

   GenericDialog gd = null;

   Vector<PlotWindow> pwv = new Vector<PlotWindow>();
   Choice    pwcw = null;
   Choice    dscw = null;
   Choice    cfcw = null;
   TextField eqsw = null;
   Vector<TextField> eqswv = new Vector<TextField>();
   Panel     eqpw = null;
   Vector<TextField> pswv = new Vector<TextField>();
   Choice    focw = null;
   Choice    fmcw = null;
   Checkbox  accw = null;
   String[]  acdf = null;
   Checkbox  afcw = null;
   Checkbox  dfcw = null;
   Checkbox  upcw = null;
   Checkbox  cccw = null;
   MultiLineLabel statusw = null;


    /** Equations of the built-in fit functions */
    public static final String[] fList = {
    "y = a+b*x",                                           // STRAIGHT_LINE
    "y = a+b*x+c*pow(x,2)",                                     // POLY2
    "y = a+b*x+c*pow(x,2)+d*pow(x,3)",                               // POLY3
    "y = a+b*x+c*pow(x,2)+d*pow(x,3)+e*pow(x,4)",                         // POLY4
    "y = a*exp(b*x)",                                      // EXPONENTIAL
    "y = a*pow(x,b)",                                           // POWER
    "y = a*log(b*x)",                                       // LOG
    "y = d+(a-d)/(1+pow(x/c,b))",                             // RODBARD
    "y = b*pow(x-a,c)*exp(-(x-a)/d)",                         // GAMMA_VARIATE
    "y = a+b*log(x-c)",                                     // LOG2
    "x = d+(a-d)/(1+pow(y/c,b)) [y = c*pow((x-a)/(d-x),1/b)]", // RODBARD2
    "y = a*exp(-b*x) + c",                                 // EXP_WITH_OFFSET
    "y = a+(b-a)*exp(-(x-c)*(x-c)/(2*d*d))",             // GAUSSIAN
    "y = a*(1-exp(-b*x)) + c",                             // EXP_RECOVERY
    "y = c*pow((x-a)/(d-x),1/b)",                           // INV_RODBARD
    "y = a*exp(b*x)",                                      // EXP_REGRESSION
    "y = a*pow(x,b)",                                           // POWER_REGRESSION
    "y = a+b*x+c*pow(x,2)+d*pow(x,3)+e*pow(x,4)+f*pow(x,5)",                   // POLY5
    "y = a+b*x+c*pow(x,2)+d*pow(x,3)+e*pow(x,4)+f*pow(x,5)+g*pow(x,6)",             // POLY6
    "y = a+b*x+c*pow(x,2)+d*pow(x,3)+e*pow(x,4)+f*pow(x,5)+g*pow(x,6)+h*pow(x,7)",       // POLY7
    "y = a+b*x+c*pow(x,2)+d*pow(x,3)+e*pow(x,4)+f*pow(x,5)+g*pow(x,6)+h*pow(x,7)+i*pow(x,8)", // POLY8
    "y = a*exp(-(x-b)*(x-b)/(2*c*c))",                    //GAUSSIAN_NOOFFSET
    "y = a*(1-exp(-b*x))",                                 //EXP_RECOVERY_NOOFFSET
    "y = a*pow(1-exp(-b*x),c)",                               //CHAPMAN
    "y = a+b*Math.erf((x-c)/d)"                                 //ERF; note that the c parameter is sqrt2 times the Gaussian
    };



   public void run(String arg) {
      int cbi = 0;
      int ci  = 0;
      int fti = 0;

      int[] ids = WindowManager.getIDList();
      if (ids == null) {
         IJ.error("No open Windows!");
         return;
         }

      gd = new NonBlockingGenericDialog("Interactive_Fitting");

      Vector<String> pwc = new Vector<String>();
      pwc.add("<choose>");
      for(int id : ids) {
         ImageWindow iw = WindowManager.getImage(id).getWindow();
         if (iw.getClass().getName().endsWith("PlotWindow")) {
            pwc.add(iw.getTitle());
            pwv.add((PlotWindow)iw);
            }
         }
      String dpwc = "";
      if (IJ.getImage().getWindow().getClass().getName().endsWith("PlotWindow"))
         dpwc = IJ.getImage().getWindow().getTitle();
      gd.addChoice("PlotWindow:",pwc.toArray(new String[0]),dpwc);
      pwcw = (Choice)gd.getChoices().get(0);
      pwcw.addItemListener(new ItemListener() { @Override public void itemStateChanged(ItemEvent e) { plotWindowChanged(); } });

      gd.addChoice("DataSource:", new String[0], "");
      dscw = (Choice)gd.getChoices().get(1);
      setDataSourceChoices();
      dscw.addItemListener(new ItemListener() { @Override public void itemStateChanged(ItemEvent e) { dataSourceChanged(); } });

      Vector<String> cfc = new Vector<String>();
      cfc.add("<choose>");
      cfc.add("<user defined>");
      for(int sti : CurveFitter.sortedTypes)
         cfc.add(CurveFitter.fitList[sti]);
      gd.addChoice("Fit:",cfc.toArray(new String[0]),"");
      cfcw = (Choice)gd.getChoices().get(2);
      cfcw.addItemListener(new ItemListener() { @Override public void itemStateChanged(ItemEvent e) { cfFunctionChanged(); } });

      gd.addStringField("Equation","",60);
      eqsw = (TextField)gd.getStringFields().get(0);
      eqsw.addActionListener(new ActionListener() { @Override public void actionPerformed(ActionEvent e) {equationChanged();} });
      eqsw.addKeyListener(new KeyListener() { 
          @Override public void keyReleased(KeyEvent e) {
             if (e.getKeyCode() == KeyEvent.VK_ENTER && e.isShiftDown())
                addEquation();
             } 
          @Override public void keyPressed(KeyEvent e) {};
          @Override public void keyTyped(KeyEvent e) {};
          });
      eqsw.setEnabled(false);
      eqsw.setVisible(false);

      eqswv.add(eqsw);

      eqpw = new Panel();
      eqpw.setLayout(new GridBagLayout());
      gd.addPanel(eqpw, GridBagConstraints.WEST, new Insets(0,110,0,0));

      gd.addMessage("Parameters to set...");

      String[] ab = "abcdefghijklmnopqrstuvwxyz".split("(?<!^)");
      for(int i=0; i<9; i++) {
         gd.addStringField(ab[i],"",60);
         pswv.add((TextField)gd.getStringFields().get(i+1));
         pswv.get(i).addActionListener(new ActionListener() { @Override public void actionPerformed(ActionEvent e) { parameterFieldChanged((TextField)e.getSource());} });
         pswv.get(i).setEnabled(false);
         pswv.get(i).setVisible(false);
         }

      Vector<String> omsv = new Vector<String>();
      omsv.add("<ignore>");
      for(int i=0; i<9; i++)
         omsv.add(ab[i]);
      gd.addChoice("Offset:",     omsv.toArray(new String[0]), "");
      focw = (Choice)gd.getChoices().get(3);
      focw.setEnabled(false);
      gd.addChoice("Multiplier:", omsv.toArray(new String[0]), "");
      fmcw = (Choice)gd.getChoices().get(4);
      fmcw.setEnabled(false);

      gd.addCheckbox("Automatically reset the parameters for doFit, as they are currently set",false); 
      accw = (Checkbox)gd.getCheckboxes().get(cbi++);
      accw.addItemListener(new ItemListener() { @Override public void itemStateChanged(ItemEvent e) { autoClearSetup(); } });

      gd.addCheckbox("Auto doFit when the identified DataSource(the plot object) changes.",false); 
      afcw = (Checkbox)gd.getCheckboxes().get(cbi++);
      afcw.addItemListener(new ItemListener() { @Override public void itemStateChanged(ItemEvent e) { autoFit(); } });

      gd.addMessage("Things to do...");

      gd.addCheckbox("doFit",false); 
      dfcw = (Checkbox)gd.getCheckboxes().get(cbi++);
      dfcw.addItemListener(new ItemListener() { @Override public void itemStateChanged(ItemEvent e) { doFit(); } });

      gd.addCheckbox("update",false);
      upcw = (Checkbox)gd.getCheckboxes().get(cbi++);
      upcw.addItemListener(new ItemListener() { @Override public void itemStateChanged(ItemEvent e) { updateFit(); } });

      gd.addCheckbox("Copy to Clipboard", false);
      cccw = (Checkbox)gd.getCheckboxes().get(cbi++);
      cccw.addItemListener(new ItemListener() { @Override public void itemStateChanged(ItemEvent e) { copyToClipboard(); } });
      
      gd.addMessage("Status to read...");
      gd.addMessage("Please select a PlotWindow.\n   Then a Plot DataSource.\n      Then an Equation.\n          Then ...");
      statusw = (MultiLineLabel)gd.getMessage();

      gd.getButtons()[0].setEnabled(false);

      gd.showDialog();
      ImagePlus.removeImageListener(this);
      if (pwcw.getSelectedIndex() == 0) return;
      if (dscw.getSelectedIndex() == 0) return;
      pwv.get(pwcw.getSelectedIndex()-1).getPlot().restorePlotObjects();
      pwv.get(pwcw.getSelectedIndex()-1).getPlot().updateImage();
      if (gd.wasCanceled()) return;
      }

   private void addEquation() {
      eqswv.add(new TextField("",60));
      GridBagConstraints c = new GridBagConstraints();
      c.gridy = eqswv.size()-1;
      c.anchor = GridBagConstraints.LINE_START;
      c.insets = new Insets(0,10,0,0);
      eqpw.add(eqswv.lastElement(),c);
      eqpw.revalidate();
      eqpw.repaint();
      gd.pack();
      }

   private void status(String str)  { status(str,0); }
   private void warning(String str) { status(str,1); }
   private void error(String str)   { status(str,2); }
   private void status(String str, int severity) {
      statusw.setText(str);
      gd.pack();
      }

   private void setDataSourceChoices() {
      if (pwcw.getSelectedIndex() > 0) {
         dscw.removeAll();
         for(String ds : pwv.get(pwcw.getSelectedIndex()-1).getPlot().getDataObjectDesignations())
            dscw.add(ds);
         dscw.insert("<choose>",0);

         pwv.get(pwcw.getSelectedIndex()-1).getPlot().savePlotObjects();
         }
      }

   private void equationChanged() {
      try {
         int np = -1;
         switch(cfcw.getSelectedIndex()) {
            case 0:  return;
            case 1:  String[] eqsf = eqsw.getText().split(";");
                     eqsw.setText(eqsf[0]);
                     np = eqsw.getText().equals("") ? -1 : CurveFitter.getNumParams(eqsw.getText());
                     if (np < 0) error("Equation bad.");
                            else status("Equations has "+np+" variables");
                     String[] vals = (eqsf.length > 1) ? eqsf[1].trim().split("\\s") : new String[0];
                     for(int i=0; i<pswv.size(); i++) {
                        pswv.get(i).setEnabled(!(i>=np));
                        pswv.get(i).setVisible(!(i>=np));
                        if (i<vals.length)
                            pswv.get(i).setText(vals[i]);
                        }      
                     //gd.revalidate();
                     gd.pack();                     
                     break;
            default: return;
            }
         }
      catch(Throwable t) { error("Equation bad."); IJ.log(""+t);}
      }
   private void parameterFieldChanged(TextField tf) {
      updateFit();
      tf.selectAll();
      }

   private void plotWindowChanged() {
      if (pwcw.getSelectedIndex() > 0) {
         setDataSourceChoices();
         gd.pack();
         }
      }

    private void dataSourceChanged() {
      if (pwcw.getSelectedIndex() == 0) return;
      if (dscw.getSelectedIndex() == 0) return;
      gd.revalidate();
      gd.pack();
      pwv.get(pwcw.getSelectedIndex()-1).getPlot().savePlotObjects();
      }

    private void cfFunctionChanged() {
      int np = -1;
      switch(cfcw.getSelectedIndex()) {
         case 0:  eqsw.setEnabled(false); 
                  eqsw.setVisible(false); 
                  focw.setEnabled(false);
                  fmcw.setEnabled(false);
                  break;
         case 1:  eqsw.setEnabled(true);  
                  eqsw.setVisible(true);
                  focw.setEnabled(true);
                  fmcw.setEnabled(true);
                  if (eqsw.getText().equals("")) {
                     status("Enter an equation of the form 'y=a*x+b' using up to parameter 'i'.\n"
                          + "Pressing the <Enter> key will enable the needed parameter field(s).\n"
                          + "A semicolon separates the equation from the white space deliminated parameter values.\n"
                          + "Pressing <shift-Enter> keys will add another equation line for plotting, not fitting.");
                     break;
                     }
                  np = CurveFitter.getNumParams(eqsw.getText());
                  status("Equation has "+np+" parameters");
                  if (np < 0) warning("Equation is probably bad.");
                  try { for(int i=0; i<np; i++) Double.parseDouble(pswv.get(i).getText()); } 
                  catch(Throwable t) { warning("Bad parameter.");}
                  break;
         default: eqsw.setEnabled(false); 
                  eqsw.setVisible(true);
                  focw.setEnabled(false);
                  fmcw.setEnabled(false);
                  eqsw.setText(/* CurveFitter. */ fList[CurveFitter.getFitCode(cfcw.getSelectedItem())]);
                  np = CurveFitter.getNumParams(CurveFitter.getFitCode(cfcw.getSelectedItem()));
                  if (!pswv.get(0).getText().equals("")) {
                     try { for(int i=0; i<np; i++) Double.parseDouble(pswv.get(i).getText()); } 
                     catch(Throwable t) { warning("Bad parameter."); }
                     }
                  status("Equation has "+np+" parameters");
                  break;
         }
      for(int i=0; i<pswv.size(); i++) {
         pswv.get(i).setEnabled(!(i>=np));
         pswv.get(i).setVisible(!(i>=np));
         }
//      cfcw.revalidate();
//      cfcw.repaint();
//      gd.revalidate();
      gd.pack();
      }

   private void autoClearSetup() {
      if (accw.getState()) {
         Vector<String> sv = new Vector<String>();
         for(TextField psw : pswv)
            if (psw.isEnabled())
               sv.add(psw.getText());
         acdf = sv.toArray(new String[0]);
         }
      else acdf = null;
      }

   private void autoFit() {
      if (afcw.getState()) ImagePlus.addImageListener(this);
                      else ImagePlus.removeImageListener(this);
      }
   public void imageOpened(ImagePlus imp)  {}
   public void imageClosed(ImagePlus imp)  { 
      if (imp == pwv.get(pwcw.getSelectedIndex()-1).getPlot().getImagePlus()) {
         status("PlotWindow closed");
         ImagePlus.removeImageListener(this);
         }
      }
   public void imageUpdated(ImagePlus imp) {
      ImagePlus.removeImageListener(this);
      if (imp == pwv.get(pwcw.getSelectedIndex()-1).getPlot().getImagePlus()) {
         pwv.get(pwcw.getSelectedIndex()-1).getPlot().savePlotObjects();
         status("Auto doFit");
         doFit();
         }
      ImagePlus.addImageListener(this);
      }

   private void doFit() {
      if (pwcw.getSelectedIndex() == 0) return;
      if (dscw.getSelectedIndex() == 0) return;
      pwv.get(pwcw.getSelectedIndex()-1).getPlot().restorePlotObjects();
      pwv.get(pwcw.getSelectedIndex()-1).getPlot().updateImage();
      float[][] data = pwv.get(pwcw.getSelectedIndex()-1).getPlot().getDataObjectArrays(dscw.getSelectedIndex()-1);
      CurveFitter cf = new CurveFitter(Tools.toDouble(data[0]), Tools.toDouble(data[1]));
      int np = -1;
      double[] p = null;
      switch(cfcw.getSelectedIndex()) {
         case 0:  return;
         case 1:  if (accw.getState()) for(int i=0; i<np; i++) pswv.get(i).setText("");
                  if (!eqsw.getText().equals("")) {
                     np = cf.getNumParams(eqsw.getText());
                     p = new double[np];
                     try { for(int i=0; i<np; i++) p[i] = Double.parseDouble(pswv.get(i).getText()); } 
                     catch(Throwable t) { error("Bad parameter.\nNo blank parameters for user function."); dfcw.setState(false); return;}
                     int fo = focw.getSelectedIndex()-1;
                     int fm = fmcw.getSelectedIndex()-1;
                     if (fo == fm && fo != -1) { error("Offset and Multiplier can not be the same.\n \n Ignoring..."); return; }
                                          else cf.setOffsetMultiplySlopeParams(fo,fm,-1);
                     try { cf.doCustomFit(eqsw.getText(),p,false); } catch(Throwable t) { error("doCustomFit: "+t); return; }
                     }
                  break;
         default: np = cf.getNumParams(cf.getFitCode(cfcw.getSelectedItem()));
                  if (acdf != null) for(int i=0; i<acdf.length; i++) pswv.get(i).setText(acdf[i]);
                  np = cf.getNumParams(cf.getFitCode(cfcw.getSelectedItem()));
                  if (!pswv.get(0).getText().equals("")) {
                     p = new double[np];
                     try { for(int i=0; i<np; i++) p[i] = Double.parseDouble(pswv.get(i).getText()); } 
                     catch(Throwable t) { error("Bad parameter."); dfcw.setState(false); return;}
                     cf.setInitialParameters(p);
                     }
                  cf.doFit(cf.getFitCode(cfcw.getSelectedItem()));
         }

      try { status(cf.getResultString()); } catch(Throwable t) { error("Fitting UGH."); return; }
      if (cf.getStatus() != Minimizer.SUCCESS) {
         dfcw.setState(false);
         return;
         }

      double[] rp = cf.getParams();
      for(int i=0; i<np; i++)
         pswv.get(i).setText(""+rp[i]);

      double[] x = cf.getXPoints();
      double[] y = cf.getYPoints();
      for(int i=0; i<x.length; i++)
         y[i] = cf.f(x[i]);

      pwv.get(pwcw.getSelectedIndex()-1).getPlot().addPoints(x,y,Plot.LINE);
      pwv.get(pwcw.getSelectedIndex()-1).getPlot().updateImage();

      gd.pack();
         
      dfcw.setState(false);
      }

   private void updateFit() {
      if (pwcw.getSelectedIndex() == 0) return;
      if (dscw.getSelectedIndex() == 0) return;
      pwv.get(pwcw.getSelectedIndex()-1).getPlot().restorePlotObjects();
      pwv.get(pwcw.getSelectedIndex()-1).getPlot().updateImage();
      int np = -1;
      double[] p = null;
      switch(cfcw.getSelectedIndex()) {
         case 0:  return;
         case 1:  if (eqsw.getText().equals("")) return;
                  np = CurveFitter.getNumParams(eqsw.getText());
                  p = new double[np];
                  try { for(int i=0; i<np; i++) p[i] = Double.parseDouble(pswv.get(i).getText()); } 
                  catch(Throwable t) { error("Bad parameter."); upcw.setState(false); return;}
                  break;
         default: np = CurveFitter.getNumParams(CurveFitter.getFitCode(cfcw.getSelectedItem()));
                  p = new double[np];
                  try { for(int i=0; i<np; i++) p[i] = Double.parseDouble(pswv.get(i).getText()); } 
                  catch(Throwable t) { error("Bad parameter."); upcw.setState(false); return;}
         }
      status("Updating...\nWith "+np+" parameters.");

      float[][] data = pwv.get(pwcw.getSelectedIndex()-1).getPlot().getDataObjectArrays(dscw.getSelectedIndex()-1);
      float[] x = data[0];
      float[] y = data[1];
      Color[] cs = new Color[]{Color.black,Color.red,Color.green,Color.blue};
      float rms = Float.NaN;
      try {
         for(int i=0; i<eqswv.size(); i++) {
            float[] yy = evaluate(eqswv.get(i).getText(),p,x);
            rms = rmse(y,yy);
            pwv.get(pwcw.getSelectedIndex()-1).getPlot().setColor(cs[i%cs.length]);
            pwv.get(pwcw.getSelectedIndex()-1).getPlot().addPoints(x,yy,Plot.LINE);
            }
         }
      catch(Throwable t) { error("One of the plots could not be made.\n"+t); gd.pack(); return;}
      pwv.get(pwcw.getSelectedIndex()-1).getPlot().updateImage();

      upcw.setState(false);
      status("Updating...\nWith "+np+" parameters.\nrmse = "+rms+"\n\nUpdated");
      gd.pack();
      }

   private float rmse(float[] a, float[] b) {
      float sum = 0F;
      for(int i=0; i<a.length; i++)
         sum += (a[i]-b[i])*(a[i]-b[i]);
      return (float)Math.sqrt(sum);
      }

   private float[] evaluate(String equation, double[] p, float[] x) {
      float[] y = new float[x.length];
      for(int i=0; i<x.length; i++)
         y[i] = evaluate(equation, p, x[i]);
      return y;
      }
   private float evaluate(String equation, double[] p, float x) {
      String code = "var x, a, b, c, d, e, f, g, h, i, j, k, l, m, n, o;\nfunction dummy() {}\n" + equation+";\n";
      Interpreter macro = new Interpreter();
      try { macro.run(code, null); } catch (Exception e) { status("Interpreter throw(ed) an exception.");   throw(e); }
      if (macro.wasError())                              { status("Interpreter.wasError() returned true."); return Float.NaN; }
      macro.setVariable("x",x);
      String[] ab = "abcdefghijklmnopqrstuvwxyz".split("(?<!^)");
      for(int i=0; i<p.length; i++)
         macro.setVariable(ab[i],p[i]);
      macro.run(39); // 21);
      return (float)macro.getVariable("y");
      }

   private void copyToClipboard() {
      java.awt.datatransfer.Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
      if (clipboard==null) { status("Could not copy to clipboard."); return; }
      String str = "";
      for(TextField tf : eqswv)
         str += tf.getText()+";\n";
      for(TextField tf : pswv)
         str += tf.getText()+"\n";
      clipboard.setContents(new StringSelection(str), null);
      status("Copied to clipboard.\n \nNB: Paste this into empty equation field and hit <enter> to repopulate all fields.");
      gd.pack();
      cccw.setState(false);
      }

   public boolean dialogItemChanged(GenericDialog gd, AWTEvent ev) {
IJ.log("dialofItemStateChanged "+ev);
      return false;
      }

   public static void arrLog(String str, float[][] arr) {
      IJ.log(str+""+arr.length+"x"+arr[0].length);
try {
      for(int i=0; i<arr[0].length; i++) {
         String s2 = "";
         for(int j=0; j<arr.length; j++)
            s2 += ""+(arr[j]!= null ? arr[j][i] : "null")+" \t ";
         IJ.log(s2);
         }
} catch(Throwable e) {IJ.log(""+e);}
      }
}
